# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.pipedag.backend.table.sql.dialects.duckdb_parquet import (
    ParquetTableStore,
)

__all__ = [
    "ParquetTableStore",
]
